from collections import namedtuple
from types import FunctionType
import dali.core as D

Beam = namedtuple("Beam", ["solution", "score", "state"])

def beam_search(initial_state,
                candidate_scores,
                make_choice,
                beam_width=5,
                eos_symbol = None,
                max_sequence_length=None,
                blacklist=[]):

    if beam_width <= 0:
        raise ValueError("Beam width must be positive, received: " + str(beam_width))

    iterations = 0
    results = [
        Beam([], D.Mat(1,1), initial_state)
    ]

    def lazy_beam(prev_beam, candidate, new_score):
        def generate():
            choice = make_choice(prev_beam.state, candidate)
            if type(choice) == tuple:
                choice_repr, new_state  = choice
            else:
                choice_repr, new_state = candidate, choice
            return Beam(
                prev_beam.solution + [choice_repr],
                new_score,
                new_state,
            )
        return generate

    def lazy_identity(beam):
        def generate():
            return beam
        return generate

    if eos_symbol is None:
        check_eos = lambda beam: False
    elif type(eos_symbol) != FunctionType:
        check_eos = lambda beam: len(beam.solution) > 0 and beam.solution[-1] == eos_symbol
    else:
        check_eos = eos_symbol

    while max_sequence_length is None or iterations < max_sequence_length:
        proposals = []
        for beam in results:
            if check_eos(beam):
                proposals.append((beam.score, lazy_identity(beam)))
            else:
                scores = candidate_scores(beam.state)
                sorted_candidates = D.MatOps.argsort(scores)

                sorted_candidates = sorted_candidates[::-1]
                candidates_remaining = beam_width
                for candidate_idx in sorted_candidates:
                    if candidate_idx in blacklist:
                        continue
                    new_score = beam.score + scores.T()[candidate_idx]
                    proposals.append((new_score, lazy_beam(beam, candidate_idx, new_score)))
                    candidates_remaining -= 1
                    if candidates_remaining <= 0:
                        break
        proposals.sort(reverse=True, key=lambda x: x[0].w[0])
        results = [ eval_beam() for _, eval_beam in proposals[:beam_width]]

        iterations += 1

    return results

__all__ = ["beam_search"]



import numpy as np
import os
import subprocess
import stat
import tempfile

from os.path import join, exists
from urllib.request import urlretrieve

from dali.core import Mat
from .misc import subsample, median_smoothing

MULTIBLEU_SCRIPT = 'multi-bleu.perl'
MULTIBLEU_URL = "https://raw.githubusercontent.com/moses-smt/mosesdecoder/master/scripts/generic/multi-bleu.perl"

class ErrorTracker(object):
    def __init__(self):
        self.epoch_error = []
        self.error_evolution = []

    def append(self, error):
        if type(error) == Mat:
            self.epoch_error.append(error.w[0,0])
        elif type(error) == np.ndarray:
            self.epoch_error.append(error[0,0])
        else:
            self.epoch_error.append(error)

    def finalize_epoch(self):
        if len(self.epoch_error) > 0:
            self.epoch_error = subsample(self.epoch_error, maximum_length=1000)
            self.error_evolution.append(self.epoch_error)
            self.epoch_error = []

    def raw(self):
        x = []
        y = []
        for epoch_no, error_epoch in enumerate(self.error_evolution):
            x.extend(float(epoch_no) + float(t) / len(error_epoch) for t in range(len(error_epoch)))
            y.extend(error_epoch)

        if len(y) > 100:
            y = median_smoothing(y, 30)
            x = x[:len(y)]
        return x,y

    def num_epochs(self):
        return len(self.error_evolution)

    def recent(self, tsteps=1):
        if len(self.epoch_error) == 0:
            return np.nan
        else:
            recent = self.epoch_error[-tsteps:]
            return sum(recent)/len(recent)

def bleu(reference, hypotheses, script_location=None):
    if script_location is None:
        script_location = join(tempfile.gettempdir(), MULTIBLEU_SCRIPT)
        if not exists(script_location):
            urlretrieve(MULTIBLEU_URL, script_location)
    else:
        assert(exists(script_location))


    def process_input(val):
        if type(val) == str:
            assert exists(val)
            return open(val)
        elif type(val) == list:
            ret = tempfile.NamedTemporaryFile("wt")
            for example in val:
                if len(example) > 0 and example[-1] == '\n':
                    ret.write(example)
                else:
                    ret.write(example + '\n')
            ret.seek(0)
            return ret

    try:
        reference  = process_input(reference)
        hypotheses = process_input(hypotheses)
        prefix = "BLEU = "

        try:
            res_str = subprocess.check_output([script_location, reference.name], stdin=hypotheses, universal_newlines=True)
        except PermissionError:
            current_permissions = os.stat(script_location).st_mode
            os.chmod(script_location, current_permissions | stat.S_IEXEC)
            res_str = subprocess.check_output([script_location, reference.name], stdin=hypotheses, universal_newlines=True)

        assert(res_str.startswith(prefix))
        res_str = res_str[len(prefix):]
        res_str = res_str.split(',')[0]
        return float(res_str)
    finally:
        if hasattr(reference, 'close'):
            reference.close()
        if hasattr(hypotheses, 'close'):
            hypotheses.close()

__all__ = [
    "ErrorTracker",
    "bleu"
]







"""
The point of this file is to make those
functions pickle friendly
"""
from dali.core import MatOps as ops

class tanh_object(object):
    def __call__(self, *args, **kwargs):
        return ops.tanh(*args, **kwargs)

class relu_object(object):
    def __call__(self, *args, **kwargs):
        return ops.relu(*args, **kwargs)

class sigmoid_object(object):
    def __call__(self, *args, **kwargs):
        return ops.sigmoid(*args, **kwargs)

class identity_object(object):
    def __call__(self, *args, **kwargs):
        assert len(args) == 1
        assert len(kwargs) == 0
        return args[0]

tanh     = tanh_object()
relu     = relu_object()
sigmoid  = sigmoid_object()
identity = identity_object()

__all__ = ["tanh", "relu", "sigmoid", "identity"]



from dali.utils import apply_recursively_on_type

class VocabEncoded(int):
    pass

class Vocab(object):
    UNK = '**UNK**'
    EOS = '**EOS**'

    def __init__(self, words=None, add_eos=True, add_unk=True):
        self.index2word = []
        self.word2index = {}
        self.eos = None
        self.unk = None
        if add_eos:
            self.add(Vocab.UNK)
        if add_unk:
            self.add(Vocab.EOS)

        if words:
            self.add(words)


    def __contains__(self, key):
        if isinstance(key, int):
            return key in range(len(self.index2word))
        elif isinstance(key, str):
            return key in self.word2index
        else:
            raise ValueError("expected(index or string)")

    def add(self, obj):
        def add_f(word):
            idx = self.word2index.get(word)
            if idx is None:
                idx = len(self.index2word)
                self.index2word.append(word)
                self.word2index[word] = idx
                if word is Vocab.UNK:
                    self.unk = idx
                if word is Vocab.EOS:
                    self.eos = idx
            return word
        apply_recursively_on_type(obj, add_f, str)

    def words(self):
        return self.word2index.keys()

    def __len__(self):
        return len(self.index2word)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.index2word[index]
        elif isinstance(index, str):
            if self.unk is not None:
                return VocabEncoded(self.word2index.get(index) or self.unk)
            else:
                return VocabEncoded(self.word2index[index])
        else:
            raise ValueError("expected(index or string)")

    def decode(self, obj, strip_eos=False, decode_type=int):
        def decode_f(word_idx):
            return self.index2word[word_idx]
        def decode_list_f(lst):
            if strip_eos:
                assert self.eos is not None
                return [el for el in lst if el != Vocab.EOS]
            else:
                return lst
        return apply_recursively_on_type(obj, decode_f, decode_type, list_callback=decode_list_f)

    def encode(self, obj, add_eos=False, encode_type=int):
        def encode_f(word):
            if self.unk is not None:
                return encode_type(self.word2index.get(word) or self.unk)
            else:
                return encode_type(self.word2index[word])
        def encode_list_f(lst):
            lst = [encode_f(word) for word in lst]
            if add_eos:
                assert self.eos is not None
                lst.append(VocabEncoded(self.eos))
            return lst
        return apply_recursively_on_type(obj, lambda x:x, str, list_callback=encode_list_f)

__all__ = [
    "VocabEncoded","Vocab"
]



import copy
from dali.core import Mat, MatOps
from dali.utils.throttled import Throttled

class SolverBase(object):
    t = Throttled(1)
    known_solvers = [
        'sgd',
        'adagrad',
        'rmsprop',
        'rmsprop_momentum',
        'adadelta',
        'adam',
    ]
    known_params = [
        'learning_rate',
        'clipval',
        'regc',
        'smooth_eps',
        'rho',
        'b1',
        'b2',
        'decay_rate',
        'gradient_normalization',
        'debug',
    ]

    def __init__(self, solver_type, **kwargs):
        self.solver_type = solver_type
        if solver_type not in SolverBase.known_solvers:
            raise AttributeError("Unknown solver " + str(solver_type))
        self.kwargs = kwargs
        for key in kwargs:
            if key not in SolverBase.known_params:
                raise AttributeError("Unknown keyword argument " + key)

    def get_arg(self, override, name, default_val):
        if name in override:
            return override[name]
        if name in self.kwargs:
            return self.kwargs[name]
        return default_val

    def param_to_cache(self, param):
        if 'solver_cache' not in param.extra_state:
            param.extra_state['solver_cache'] = {}
        return param.extra_state['solver_cache']

    def step(self, params, param_caches=None, **kwargs_override):
        if type(params) != list:
            params = [params]
            if param_caches is not None:
                param_caches = [param_caches]

        if param_caches is None:
            param_caches = [None for _ in range(params)]
        assert len(params) == len(param_caches)

        for key in kwargs_override:
            if not key in SolverBase.known_params:
                raise AttributeError("Unknown keyword argument " + key)

        debug = self.get_arg(kwargs_override, 'debug', ['nans'])
        clip_val = self.get_arg(kwargs_override, 'clipval', 5.0)
        regc     = self.get_arg(kwargs_override, 'regc',     0.0)

        ########## GRADIENT NORMALIZATION ###########

        gradient_normalization = self.get_arg(kwargs_override, 'gradient_normalization', 'norm')
        if gradient_normalization == 'norm':
            for param in params:
                if regc > 0.0:
                    MatOps.regularize(param, regc)
                MatOps.normalize(param, clip_val)
        elif gradient_normalization == 'clipping':
            for param in params:
                MatOps.clip_and_regularize(param, clip_val, regc)
        elif gradient_normalization == 'discard':
            params_exceeding = []
            for param in params:
                if MatOps.grad_norm(param).w[0,0] > clip_val:
                    params_exceeding.append(param.name if param.name != '' else '(unnamed parameter)')
            if len(params_exceeding) > 0:
                if 'discards' in debug:
                    print('Discarding gradient update due to exceeded norm for the following parameters: %s' % (params_exceeding,))
                for param in params:
                    param.clear_grad()
                return
        elif gradient_normalization == 'none':
            if regc > 0.0:
                MatOps.regularize(param, regc)
        else:
            raise AttributeError("Unknown gradient_normalization mode : " + gradient_normalization)


        ########## SOLVING ###########

        for param, param_cache in zip(params, param_caches):
            if param_cache is None:
                param_cache = self.param_to_cache(param)

            learning_rate = self.get_arg(kwargs_override, "learning_rate", 0.01)

            lr_multiplier = param.extra_state.get('lr_multiplier', 1.0)
            learning_rate *= lr_multiplier
            if MatOps.is_grad_nan(param):
                if SolverBase.t.should_i_run() and 'nans' in debug:
                    name_str = ' (unnamed parameter)'
                    if param.name is not None:
                        name_str = ' (name: %s)' % (param.name,)
                    print("Warning ignoring grad update due to NaNs%s." % (name_str,))
            else:
                if self.solver_type == 'sgd':
                    MatOps.sgd_update(param, learning_rate)
                elif self.solver_type == 'adagrad':
                    smooth_eps = self.get_arg(kwargs_override, "smooth_eps", 1e-6)
                    cache = self.get_cache(param, param_cache, 'adagrad_cache')
                    MatOps.adagrad_update(param, cache, learning_rate, smooth_eps)
                elif self.solver_type == 'rmsprop':
                    smooth_eps = self.get_arg(kwargs_override, "smooth_eps", 1e-6)
                    decay_rate = self.get_arg(kwargs_override, "decay_rate", 0.95)
                    cache = self.get_cache(param, param_cache, 'rmsprop_cache')
                    MatOps.rmsprop_update(param, cache, decay_rate, learning_rate, smooth_eps)
                elif self.solver_type == 'rmsprop_momentum':
                    decay_rate = self.get_arg(kwargs_override,    "decay_rate", 0.95)
                    momentum = self.get_arg(kwargs_override,      "momentum",   0.9)
                    smooth_eps = self.get_arg(kwargs_override,    "smooth_eps", 1e-4)
                    n_cache = self.get_cache(param, param_cache, 'rmsprop_momentum_n_cache')
                    g_cache = self.get_cache(param, param_cache, 'rmsprop_momentum_g_cache')
                    momentum_cache = self.get_cache(param, param_cache, 'rmsprop_momentum_momentum_cache')
                    MatOps.rmsprop_momentum_update(param, n_cache, g_cache, momentum_cache, decay_rate, momentum, learning_rate, smooth_eps)
                elif self.solver_type == 'adadelta':
                    smooth_eps = self.get_arg(kwargs_override, "smooth_eps", 1e-4)
                    rho        = self.get_arg(kwargs_override, "rho",        0.95)
                    gsum = self.get_cache(param, param_cache, 'adadelta_gsum')
                    xsum = self.get_cache(param, param_cache, 'adadelta_xsum')
                    MatOps.adadelta_update(param, gsum, xsum, rho, smooth_eps)
                elif self.solver_type == 'adam':
                    smooth_eps = self.get_arg(kwargs_override, "smooth_eps", 1e-4)
                    b1         = self.get_arg(kwargs_override, "b1",        0.5)
                    b2         = self.get_arg(kwargs_override, "b2",        1e-6)
                    m  = self.get_cache(param, param_cache, 'adam_m')
                    v  = self.get_cache(param, param_cache, 'adam_v')
                    epoch = param.extra_state.get('adam_epoch', 1)

                    MatOps.adam_update(param, m, v, b1, b2, smooth_eps, learning_rate, epoch)

                    param_cache['adam_epoch'] = epoch + 1
                else:
                    assert False
            param.clear_grad()

    def set_lr_multiplier(self, param, lr_multiplier):
        param.extra_state["lr_multiplier"] = lr_multiplier

    def get_cache(self, param, cache_state, cache_name):
        if cache_name not in cache_state:
            cache_state[cache_name] = Mat.zeros(param.shape, dtype=param.dtype)
        ret = cache_state[cache_name]
        assert ret.shape == param.shape, \
                "Wrong parameter passed to solver (cache shape does not match parameter's shape)"
        return ret

    def reset_caches(self, param, param_caches=None):
        if type(param) == list:
            if param_caches is not None:
                assert len(param) == len(param_caches)
                for p,c  in zip(param, param_caches):
                    self.reset_caches(p, c)
            else:
                for p in param:
                    self.reset_caches(p)
        elif type(param) == Mat:
            # get caches
            if param_caches is None:
                param_caches = self.param_to_cache(param)
            # reset
            if self.solver_type == 'adagrad':
                self.get_cache(param, param_caches, 'adagrad_cache').clear()
            elif self.solver_type == 'rmsprop':
                self.get_cache(param, param_caches, 'rmsprop_cache').clear()
            elif self.solver_type == 'adadelta':
                self.get_cache(param, param_caches, 'adadelta_gsum').clear()
                self.get_cache(param, param_caches, 'adadelta_xsum').clear()
            elif self.solver_type == 'adam':
                self.get_cache(param, param_caches, 'adam_m').clear()
                self.get_cache(param, param_caches, 'adam_v').clear()
                if 'adam_epoch' in param:
                    del param_caches["adam_epoch"]
            else:
                assert False

class Solver(object):
    def __init__(self, parameters, *args, **kwargs):
        """Solver

        It is pickleable.
        """
        self.base = SolverBase(*args, **kwargs)

        if type(parameters) == list:
            self._parameters = parameters
        else:
            self._parameters = [parameters]
        self.caches =  [{} for _ in range(len(self._parameters))]

        self.lr_multipliers = [None for _ in range(len(self._parameters))]

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, val):
        assert self._parameters is None or \
                len(self._parameters) == len(val), \
                "Number of parameters must remain unchanged"
        self._parameters = val
        for lr_multiplier, param in zip(self.lr_multipliers, self._parameters):
            if lr_multiplier is not None:
                self.base.set_lr_multiplier(param, lr_multiplier)

    def set_lr_multiplier(self, where, val):
        indices = []
        if type(where) == str:
            for i, param in enumerate(self.parameters):
                if param.name == where:
                    indices.append(i)
            assert len(indices) > 0, \
                    "Could not find parameters %s" % (where,)
        elif type(where) == int:
            indices.append(where)
        else:
            raise ValueError("where must be int or str")
        for i in indices:
            self.lr_multipliers[i] = val
            self.base.set_lr_multiplier(self.parameters[i], val)

    def step(self):
        assert self.parameters is not None, \
                "Remeber to use set parameters after unpickling."
        self.base.step(self.parameters, self.caches)

    def reset_caches(self):
        assert self.parameters is not None, \
                "Remeber to use set parameters after unpickling."
        self.base.reset_caches(self.parameters, self.caches)

    @property
    def solver_type(self):
        return self.base.solver_type

    def __setstate__(self, state):
        self.base          = state['solver']
        self.caches       = state['caches']
        self.lr_multipliers = state['lr_multipliers']
        self._parameters = None

    def __getstate__(self):
        return {
            'solver'      : self.base,
            'caches'      : self.caches,
            'lr_multipliers': self.lr_multipliers
        }

class CombinedSolver(object):
    def __init__(self, solvers):
        self.solvers = solvers

    def step(self):
        for solver in self.solvers:
            solver.step()

    def reset_caches(self):
        for solver in self.solvers:
            solver.reset_caches()

__all__ = [
    "CombinedSolver",
    "SolverBase",
    "Solver",
]



from collections import defaultdict

from .batch import TranslationBatch
from .process import Process, Multiplexer, Lines, DiscoverFiles, IdentityReducer, BatchBenefactor
from dali.utils import Vocab


class TranslationFiles(object):
    def __init__(self, root_path, from_lang, to_lang):
        files_from = set(DiscoverFiles(root_path, "." + from_lang))
        files_to   = set(DiscoverFiles(root_path, "." + to_lang))

        self.pairs = []
        for file_name in files_from:
            pref = file_name[:-(len(from_lang) + 1)]
            hypothetical_to_file = pref + '.' + to_lang
            if hypothetical_to_file in files_to:
                self.pairs.append((file_name, hypothetical_to_file))

        self.next_pair = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.next_pair >= len(self.pairs):
            raise StopIteration()
        else:
            self.next_pair += 1
            return self.pairs[self.next_pair - 1]


def TranslationMapper(sentence_bounds=(None, None)):
    def translation_lines():
        lines =  Lines()                         \
                 .split_punctuation()            \
                 .split_spaces()                 \
                 .bound_length(*sentence_bounds)
        return lines


    return Multiplexer(translation_lines(), translation_lines())


def build_vocabs(path, from_lang, to_lang, from_max_size=None, to_max_size=None):
    from_occurence = defaultdict(lambda: 0)
    to_occurence   = defaultdict(lambda: 0)
    try:
        for from_sentence, to_sentence in Process(files=TranslationFiles(path, from_lang, to_lang),
                                                  mapper=TranslationMapper(sentence_bounds=(None, None)),
                                                  reducer=IdentityReducer()):
            for word in from_sentence:
                from_occurence[word] += 1

            for word in to_sentence:
                to_occurence[word] += 1
    except KeyboardInterrupt:
        print('Impatient User Detected, file processing halted, proceeding to build vocab.')


    from_occurence = list(from_occurence.items())
    to_occurence   = list(to_occurence.items())

    # highest occurrence first
    from_occurence.sort(key=lambda x: x[1], reverse=True)
    to_occurence  .sort(key=lambda x: x[1], reverse=True)

    # remove occurences, keep sorted words
    from_occurence = [x[0] for x in from_occurence]
    to_occurence =   [x[0] for x in to_occurence]

    from_vocab = Vocab(from_occurence[:from_max_size])
    to_vocab   = Vocab(to_occurence[:to_max_size])

    return from_vocab, to_vocab

def iterate_examples(root_path, from_lang, to_lang, vocabs, minibatch_size, reverse_input=True, sentences_until_minibatch=None, sentence_length_bounds=(None, None)):
    sentences_until_minibatch = sentences_until_minibatch or 10000 * minibatch_size
    files   = TranslationFiles(root_path, from_lang, to_lang)
    mapper = TranslationMapper(sentence_bounds=sentence_length_bounds)
    sorting_key = lambda sentence_pair: (len(sentence_pair[0]), len(sentence_pair[1])) # sort by length of the input sentence first and then by the length of the output sentence

    reducer = BatchBenefactor(minibatch_size,
                              TranslationBatch.given_vocabs(vocabs, store_originals=True, reverse_input=reverse_input),
                              sentences_until_minibatch,
                              sorting_key=sorting_key)
    return Process(files=files, mapper=mapper, reducer=reducer)



import os
import random

from queue import Queue

from dali.data.utils import split_punctuation as split_punctuation_f

class Process(object):
    def __init__(self, files, mapper, reducer):
        if files == str:
            files = (make_me_iterator for make_me_iterator in [files])
        self.files   = files
        self.mapper  = mapper
        self.reducer = reducer

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            try:
                return next(self.reducer)
            except StopIteration:
                pass
            # if we got here it means reducer run out of elements
            # and we need to add more from mapper.
            try:
                self.reducer.add(next(self.mapper))
                continue
            except StopIteration:
                pass
            # if we got here it means that mapper run out of elements and we need to give
            # if another file
            next_file = next(self.files)
            self.mapper.set_file(next_file)

class DiscoverFiles(object):
    def __init__(self, root_path, extension=None):
        self.files = []
        self.next_file = 0
        for path, dirs, files in os.walk(root_path):
            for file in files:
                if extension is None or file.endswith(extension):
                    self.files.append(os.path.join(path, file))

    def __iter__(self):
        return self

    def __next__(self):
        if self.next_file >= len(self.files):
            raise StopIteration()
        else:
            self.next_file += 1
            return self.files[self.next_file - 1]


class FileMapper(object):
    FILTER      = 1
    TRANSFORMER = 2

    def __init__(self):
        self._transformations = []
        self.file_name = None
        self.file_handle = None

    def set_file(self, file_name):
        if self.file_handle is not None:
            self.file_handle.close()
            self.file_handle = None
        self.file_name = file_name
        self.file_handle = None


    def get_file(self, fargs="rt"):
        if self.file_handle is None:
            self.file_handle = open(self.file_name, fargs)
        return self.file_handle

    def __del__(self):
        if self.file_handle is not None:
            self.file_handle.close()
            self.file_handle = None

    def __next__(self):
        if self.file_name is None:
            raise StopIteration()
        while True:
            item = self.next_item_no_filter()
            if item is not None:
                return item

    def next_item_no_filter(self):
        """By default __next__ does not even return an item
           if it does not pass a filer. This function, returns
           None even if the filters are not passing. This
           is useful for example for translation where,
           we want to ensure that two different file streams
           are always aligned
        """
        if self.file_name is None:
            raise StopIteration()

        item = self.next_item()
        item = self.transform_item(item)

        return item

    def transform_item(self, item):
        for transform_f in self._transformations:
            item = transform_f(item)
            if item is None:
                break
        return item

    def next_item(self):
        raise StopIteration()


    def add_filter(self, filter_f):
        def wrapper(element):
            if filter_f(element):
                return element
            return None
        self.add_transform(wrapper)
        return self

    def add_transform(self, transform_f):
        self._transformations.append(transform_f)
        return self

    def __iter__(self):
        return self

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state['file_handle']
        if self.file_handle is not None:
            state['__file_position'] = self.file_handle.tell()
        return state

    def __setstate__(self, state):
        # Restore instance attributes (i.e., filename and lineno).
        self.__dict__.update(state)
        # Restore the previously opened file's state. To do so, we need to
        # reopen it and read from it until the line count is restored.
        self.file_handle = open(self.file_name)
        if '__file_position' in state:
            self.file_handle.seek(state['__file_position'])


class Lines(FileMapper):
    def next_item(self):
        res = self.get_file().readline()
        if len(res) == 0:
            raise StopIteration()
        if len(res) > 0 and res[-1] == '\n':
            return res[:-1]
        else:
            return res

    def lower(self):
        return self.add_transform(lambda x: x.lower())

    def bound_length(self, lower_bound=None, upper_bound=None):
        if lower_bound:
            self.add_filter(lambda x: lower_bound <= len(x))
        if upper_bound:
            self.add_filter(lambda x: len(x) <= upper_bound)
        return self

    def split_spaces(self):
        return self.add_transform(lambda x: x.split(' '))

    def reverse(self):
        return self.add_transform(lambda x: list(reversed(x)))

    def split_punctuation(self):
        return self.add_transform(split_punctuation_f)


class Multiplexer(object):
    def __init__(self, *mappers):
        self.mappers = mappers

    def set_file(self, args):
        assert len(args) == len(self.mappers)

        for mapper, arg in zip(self.mappers, args):
            mapper.set_file(arg)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            ret = tuple([m.next_item_no_filter() for m in self.mappers])
            if all(ret_i is not None for ret_i in ret):
                return ret



class BatchBenefactor(object):
    def __init__(self, minibatch_size,
                 minibatch_class=None,
                 examples_until_minibatches=None,
                 sorting_key=None):
        self.minibatch_size = minibatch_size
        self.minibatch_class = minibatch_class
        self.examples_until_minibatches = examples_until_minibatches or minibatch_size
        self.sorting_key = sorting_key

        self.collected = []
        self.batches = []
        self.next_batch = 0

    def add(self, element):
        self.collected.append(element)

        if len(self.collected) >= self.examples_until_minibatches:
            self.batches = []
            self.next_batch = 0

            sorting_key = self.sorting_key or (lambda x: len(x))
            self.collected.sort(key=sorting_key)


            batch_start_idxes = list(range(0, len(self.collected), self.minibatch_size))
            random.shuffle(batch_start_idxes)
            for i in batch_start_idxes:
                if i + self.minibatch_size <= len(self.collected):
                    self.batches.append(self.collected[i:(i + self.minibatch_size)])
            self.collected = []

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.batches) == 0 or self.next_batch == len(self.batches):
            self.batches = []
            self.next_batch = 0
            raise StopIteration()
        else:
            assert self.next_batch < len(self.batches)
            self.next_batch += 1
            minibatch_class = self.minibatch_class or (lambda x:x)
            return minibatch_class(self.batches[self.next_batch - 1])



    def update_minibatch_size(self, minibatch_size):
        self.minibatch_size = minibatch_size


class IdentityReducer(object):
    def __init__(self):
        self.q = Queue()

    def add(self, element):
        self.q.put(element)

    def __iter__(self):
        return self

    def __next__(self):
        if self.q.empty():
            raise StopIteration()
        else:
            return self.q.get()



import numpy as np

import dali.core as D

class Batch(object):
    def __init__(self):
        self.timesteps = 0
        self.examples  = 0
    def inputs(timestep):
        return None
    def targets(timestep):
        return None
    def __repr__(self):
        return 'Batch(timesteps=%d, examples=%d)' % (self.timesteps, self.examples)

START_TOKEN = '**START**'

def create_lines_batch(lines, vocab, add_start_token=False, fill_eos=False, add_eos=False, align_right=False):
    encoded_lines = []
    for l in lines:
        if type(l) == str:
            l = l.split(" ")
        if add_start_token:
            l = [START_TOKEN] + l
        encoded_lines.append(vocab.encode(l, add_eos=add_eos))


    seq_length = max(map(len, encoded_lines))
    # we add one index to account for start of sequence token
    data = np.empty((seq_length, len(lines)), dtype=np.int32)
    if fill_eos:
        data.fill(vocab.eos)
    else:
        data.fill(0)

    for line_idx, encoded_line in enumerate(encoded_lines):
        if align_right:
            data[-len(encoded_line):, line_idx] = encoded_line
        else:
            data[:len(encoded_line), line_idx] = encoded_line
    data = D.Mat(data, borrow=True, dtype=np.int32)
    return data

class LMBatch(object):
    @staticmethod
    def given_vocab(vocab, **kwargs):
        def wrapper(sentences):
            return LMBatch(sentences, vocab, **kwargs)
        return wrapper

    def __init__(self, sentences, vocab, store_originals=False, fill_eos=True, add_eos=True, add_start_token=True):
        if store_originals:
            self.sentences = sentences
        self.sentence_lengths = [len(s) for s in sentences]
        self.data = create_lines_batch(
            sentences,
            vocab,
            add_start_token=add_start_token,
            fill_eos=fill_eos,
            add_eos=add_eos
        )
        self.timesteps = self.data.shape[0] - 1
        self.examples  = self.data.shape[1]

    def inputs(self, timestep):
        return self.data[timestep]

    def targets(self, timestep):
         # predictions are offset by 1 to inputs, so
        return self.data[timestep + 1]


class TranslationBatch(object):
    @staticmethod
    def given_vocabs(vocabs, **kwargs):
        def wrapper(sentence_pairs):
            return TranslationBatch(sentence_pairs, vocabs, **kwargs)
        return wrapper


    def __init__(self, sentence_pairs, vocabs, store_originals=False, input_add_eos=False, output_add_eos=True, add_start_token=False, reverse_input=True):
        if store_originals:
            self.sentence_pairs = sentence_pairs
        from_sentences = [sentence_pair[0] for sentence_pair in sentence_pairs]
        to_sentences   = [sentence_pair[1] for sentence_pair in sentence_pairs]

        if reverse_input:
            from_sentences = [list(reversed(s)) for s in from_sentences]


        from_vocab, to_vocab = vocabs

        from_eos_correction = (1 if input_add_eos else 0)
        to_eos_correction   = (1 if output_add_eos else 0)

        self.from_data = create_lines_batch(
            from_sentences,
            from_vocab,
            add_start_token=add_start_token,
            fill_eos=False,
            add_eos=input_add_eos,
            align_right=True
        )

        self.to_data = create_lines_batch(
            to_sentences,
            to_vocab,
            add_start_token=add_start_token,
            fill_eos=True,
            add_eos=output_add_eos
        )
        self.from_tokens  = sum(map(len, from_sentences)) + from_eos_correction * len(from_sentences)
        self.to_tokens    = sum(map(len, to_sentences))   + to_eos_correction  * len(to_sentences)

        self.target_mask = D.Mat(*self.to_data.shape)
        for example_idx, sentence in enumerate(to_sentences):
            for ts in range(len(sentence) + to_eos_correction):
                self.target_mask.w[ts, example_idx] = 1
                self.target_mask.constant = True

        self.from_timesteps = self.from_data.shape[0]
        self.timesteps = self.from_data.shape[0] + self.to_data.shape[0]
        self.examples  = len(sentence_pairs)

    def inputs(self, timestep):
        if timestep < self.from_timesteps:
            return self.from_data[timestep]
        else:
            return None

    def targets(self, timestep):
        if timestep >= self.from_timesteps:
            return self.to_data[timestep - self.from_timesteps]
        else:
            return None

    def masks(self, timestep):
        if timestep >= self.from_timesteps:
            return self.target_mask[timestep - self.from_timesteps]
        else:
            return None



from collections import defaultdict

class Capture(object):
    instances = set()

    def __init__(self):
        self.state = defaultdict(lambda: [])

    @classmethod
    def add(cls, name, value):
        for instance in cls.instances:
            instance.state[name].append(value)

    def __enter__(self):
        Capture.instances.add(self)

    def __exit__(self, *args, **kwargs):
        Capture.instances.remove(self)



import dali.core as D

class MLP(object):
    def __init__(self, input_sizes, hiddens, nonlinearities):
        self.input_sizes = input_sizes
        self.hiddens = hiddens
        self.input_nonlinearity, self.layer_nonlinearities = nonlinearities[0], nonlinearities[1:]

        self.input_layer = D.StackedInputLayer(input_sizes, hiddens[0])
        self.layers = [D.Layer(h_from, h_to) for h_from, h_to in zip(hiddens[:-1], hiddens[1:])]

    def activate(self, inputs):
        assert len(self.layers) == len(self.layer_nonlinearities)
        hidden = self.input_nonlinearity(self.input_layer.activate(inputs))
        for l, nonlinearity in zip(self.layers, self.layer_nonlinearities):
            hidden = nonlinearity(l.activate(hidden))
        return hidden

    def parameters(self):
        ret = self.input_layer.parameters()
        for l in self.layers:
            ret.extend(l.parameters())
        return ret

    def name_parameters(self, prefix):
        self.input_layer.name_parameters(prefix + "_input_layer")
        for layer_idx, layer in enumerate(self.layers):
            layer.name_parameters(prefix + '_layer%d' % (layer_idx,))



import subprocess
import preprocessor
import distutils.ccompiler
import distutils.sysconfig
import subprocess

from os.path import join, dirname, realpath, exists, getmtime
from os      import environ, walk, makedirs
from sys import platform
import numpy as np

from distutils.core import setup
from distutils.command import build as build_module, clean as clean_module
from Cython.Distutils.extension import Extension
from Cython.Distutils import build_ext

from tempfile import TemporaryDirectory

SCRIPT_DIR = dirname(realpath(__file__))
DALI_CORE_DIR = join(SCRIPT_DIR, "dali", "core")

def find_extension_files(path, extension):
    """Recursively find files with specific extension in a directory"""
    for relative_path, dirs, files in walk(path):
        for fname in files:
            if fname.endswith(extension):
                yield join(path, relative_path, fname)


def cmake_robbery(varnames, fake_executable="dummy"):
    """Capture Cmake environment variables by running `find_package(dali)`"""
    varstealers = []
    magic_command = "CYTHON_DALI_BEGIN_VARIABLE_STEALING"
    varstealers.append("message(STATUS \"%s\")" % (magic_command,))
    for varname in varnames:
        varstealers.append("message(STATUS  \"CYTHON_DALI_%s: ${%s}\")" % (varname, varname,))
    varstealers = "\n".join(varstealers) + "\n"

    with TemporaryDirectory() as temp_dir:
        with open(join(temp_dir, "source.cpp"), "wt") as source_cpp:
            source_cpp.write("int main() {};\n")
        with open(join(temp_dir, "CMakeLists.txt"), "wt") as cmake_conf:
            cmake_conf.write("""
                cmake_minimum_required(VERSION 2.8 FATAL_ERROR)
                project("dali-cython")
                find_package(Dali REQUIRED) # find Dali.
                add_executable(%s source.cpp)
                target_link_libraries(%s ${DALI_LIBRARIES})
            """ % (fake_executable, fake_executable,) + varstealers)

        cmake_subdirectory = fake_executable + ".dir"
        cmake_stdout = subprocess.check_output(
            ["cmake", "."], cwd=temp_dir, universal_newlines=True, stderr=subprocess.DEVNULL
        )

        # capture the link arguments
        with open(join(temp_dir, "CMakeFiles", cmake_subdirectory, "link.txt"), "rt") as f:
            linking_command = f.read()

    linking_args = linking_command.split(" ", 1)[1].strip().split()
    linking_args = [arg for arg in linking_args if cmake_subdirectory not in arg]
    outvars = {}
    outvars["LINK_ARGS"] = linking_args

    # slice output after the magic command and retrieve these variables
    # from the CMake environment
    idx = cmake_stdout.find(magic_command) + len(magic_command) + 1
    lines = cmake_stdout[idx:].split("\n")[:len(varnames)]

    for varname, line in zip(varnames, lines):
        assert(varname in line)
        _, value = line.split(":", 1)
        outvars[varname] = value.strip().split(";")
    return outvars

# cmake environment variables
robbed = cmake_robbery(["DALI_INCLUDE_DIRS"])

# set the compiler
if platform == 'linux':
    environ["cc"] = 'gcc'
    environ["cc"] = 'g++'
else:
    environ["CC"] = "clang"
    environ["CXX"] = "clang++"

# Make a `cleanall` rule to get rid of intermediate and library files
class clean(clean_module.clean):
    def run(self):
        print("Cleaning up cython files...")
        # Just in case the build directory was created by accident,
        # note that shell=True should be OK here because the command is constant.
        for place in ["build", "dali/core.c", "dali/core.cpp", "dali/*.so"]:
            subprocess.Popen("rm -rf %s" % (place,), shell=True, executable="/bin/bash", cwd=SCRIPT_DIR)

compiler = distutils.ccompiler.new_compiler()
distutils.sysconfig.customize_compiler(compiler)
BLACKLISTED_COMPILER_SO = ['-Wp,-D_FORTIFY_SOURCE=2']
build_ext.compiler = compiler

ext_modules = [Extension(
    name='dali.core',
    sources=[join(SCRIPT_DIR, "dali/core.pyx")] + list(find_extension_files(DALI_CORE_DIR, ".cpp")),
    library_dirs=[],
    language='c++',
    extra_compile_args=['-std=c++11'],
    extra_link_args=robbed["LINK_ARGS"],
    libraries=[],
    extra_objects=[],
    include_dirs=[np.get_include()] + robbed["DALI_INCLUDE_DIRS"]
)]

def run_preprocessor():
    """
    Generate python files using a file prepocessor (essentially macros
    that generate multiple versions of the code for each dtype supported
    by a Dali operation)
    """
    EXTENSION = ".pre"
    for py_processor_file in find_extension_files(SCRIPT_DIR, EXTENSION):
        output_file = py_processor_file[:-len(EXTENSION)]

        if not exists(output_file) or \
                getmtime(py_processor_file) > getmtime(output_file) or \
                getmtime(join(SCRIPT_DIR, "preprocessor_utils.py")) > getmtime(output_file):
            print('Preprocessing %s' % (py_processor_file,))
            with open(output_file, "wt") as f:
                f.write(preprocessor.process_file(py_processor_file, prefix='pyp', suffix='ypy'))

# We need to remove some compiler flags, to make sure
# the code can compile on Fedora (to be honest it seems
# to be a bug in Fedora's distrubtion of Clang).
# Nevertheless this little madness below is to change
# default compiler flags used by Cython.
# If you know a better way call me immediately day
# or night at 4getszymo4. Thank you!
class nonbroken_build_ext(build_ext):
    def build_extensions(self, *args, **kwargs):
        run_preprocessor()
        new_compiler_so = []
        for arg in self.compiler.compiler_so:
            if arg not in BLACKLISTED_COMPILER_SO:
                new_compiler_so.append(arg)
        self.compiler.compiler_so = new_compiler_so
        super(nonbroken_build_ext, self).build_extensions(*args, **kwargs)

setup(
  name="dali",
  version='1.0.0',
  cmdclass={"build_ext": nonbroken_build_ext, 'clean': clean},
  ext_modules=ext_modules,
  description="Buttery smooth automatic differentiation using Dali.",
  author="Jonathan Raiman, Szymon Sidor",
  author_email="jonathanraiman at gmail dot com",
  install_requires=[
    'preprocessor',
    'numpy',
    'dill'
  ],
  packages=[
    "dali",
    "dali.utils",
    "dali.data",
    "dali.models",
  ]
)



import dill as pickle
import inspect
import numpy as np
import types

from os import makedirs, listdir
from os.path import join, exists

import dali.core as D

class RunningAverage(object):
    def __init__(self, alpha=0.95):
        self.alpha = alpha
        self.value = None

    def update(self, measurement):
        if self.value is None:
            self.value = measurement
        else:
            self.value = (self.alpha * self.value +
                         (1.0 - self.alpha) * measurement)

    def __float__(self):
        return float(self.value)


def apply_recursively_on_type(x, f, target_type, list_callback=None):
    if type(x) == target_type:
        return f(x)
    elif type(x) == list or isinstance(x, types.GeneratorType):
        ret = [ apply_recursively_on_type(el, f, target_type, list_callback) for el in x]
        if list_callback and all(type(el) == target_type for el in x):
            ret = list_callback(ret)
        return ret
    elif type(x) == dict:
        res = {}
        for k,v in x.items():
            res[k] = apply_recursively_on_type(v, f, target_type, list_callback)
        return res
    else:
        return x

def integer_ceil(a, b):
    return (a + b - 1) // b

def subsample(seq, maximum_length):
    if seq == []:
        return seq
    return seq[::integer_ceil(len(seq), maximum_length)]

def median_smoothing(signal, window=10):
    res = []
    for i in range(window, len(signal)):
        actual_window = signal[i-window:i]
        res.append(np.median(actual_window))
    return res

def pickle_from_scope(directory, variables, caller_globals=None, caller_locals=None):
    if not exists(directory):
        makedirs(directory)

    if caller_globals is None or caller_locals is None:
        stack = inspect.stack()
        if caller_globals is None:
            caller_globals = stack[1][0].f_globals
        if caller_locals is None:
            caller_locals  = stack[1][0].f_locals
        del stack

    for var in variables:
        with open(join(directory, var + ".pkz"), "wb") as f:
            value = caller_locals.get(var) or caller_globals.get(var)
            assert value is not None
            pickle.dump(value, f)

def unpickle_as_dict(directory, whitelist=None, extension='.pkz'):
    assert exists(directory)

    res = {}

    for file_name in listdir(directory):
        if file_name.endswith(extension):
            var_name = file_name[:-len(extension)]
            if whitelist is None or var_name in whitelist:
                with open(join(directory, file_name), "rb") as f:
                    res[var_name] = pickle.load(f)

    return res

def add_device_args(parser):
    parser.add_argument("--device",    type=str, default='gpu', choices=['gpu','cpu'], help="Whether model should run on GPU or CPU.")
    parser.add_argument("--gpu_id",    type=int, default=0, help="Which GPU to use (zero-indexed just like in CUDA APIs)")

def set_device_from_args(args, verbose=False):
    D.config.default_device = args.device
    if args.device == 'gpu':
        D.config.default_gpu = args.gpu_id
        if verbose:
            print("Using %s" % (D.config.gpu_id_to_name(args.gpu_id)))

__all__ = [
    "apply_recursively_on_type",
    "integer_ceil",
    "subsample",
    "median_smoothing",
    "pickle_from_scope",
    "unpickle_as_dict",
    "RunningAverage",
    "add_device_args",
    "set_device_from_args"
]



import time




class Throttled(object):
    def __init__(self, min_time_since_last_run_s=5):
        """Used for simple throttled execution.

        Here's a simple example:

            @Throttled(1)
            def lol(i):
                print('epoch %d' % (i,), flush=True)

            for i in range(100000000):
                lol(i)

        Above code will report the epoch every second.

        Here's another way:

            throttled = Throttled(1)


            for i in range(100000000000):
                if throttled.should_i_run():
                    print('epoch %d' % (i,), flush=True)
        """
        self.last_time = None
        self.min_time_since_last_run_s = min_time_since_last_run_s


    def should_i_run(self, min_time_since_last_run_s=None):
        min_time_since_last_run_s = min_time_since_last_run_s or self.min_time_since_last_run_s
        now = time.time()
        if self.last_time is None or (now - self.last_time) > min_time_since_last_run_s:
            self.last_time = now
            return True
        else:
            return False

    def maybe_run(self, f, min_time_since_last_run_s=None):
        if self.should_i_run(min_time_since_last_run_s):
            return f()
        else:
            return None

    def __call__(self, f):
        def wrapper(*args, **kwargs):
            return self.maybe_run(lambda: f(*args, **kwargs))
        return wrapper



PUNCTUATION_CHARS = set(list('.,?!-"\'()[]{}:;'))

def split_punctuation(sentence, punctuation=PUNCTUATION_CHARS):
    res = []
    for i, char in enumerate(list(sentence)):
        if char in punctuation:
            if i - 1 >= 0 and sentence[i-1] != ' ':
                res.append(' ')
            res.append(char)
            if i + 1 < len(sentence) and sentence[i + 1] != ' ':
                res.append(' ')
        else:
            res.append(char)
    return ''.join(res)
